from ..client import ClientThread
from ..client import create_to_path


def do_put_file(args):
    client = ClientThread(args.uri)
    client.start()
    remotefile = create_to_path(args.localfile, args.remotefile)
    client.put_file(args.localfile, remotefile)


def add_subparser(subparsers):
    subparser = subparsers.add_parser('put_file')
    subparser.add_argument('-u' ,'--uri',
                           default='tcp://127.0.0.1:28000',
                           help='URI of the server (default: %(default)s)')
    subparser.add_argument('localfile', help='The local file path.')
    subparser.add_argument('remotefile',
                           nargs='?',
                           help='The remote file path.')
    subparser.set_defaults(func=do_put_file)



import asyncio
import sys
import os
import threading
import logging
import time
import re

from colors import red
from colors import yellow
from colors import green
from colors import cyan

from .version import __version__
from .bunga_client import BungaClient


LOGGER = logging.getLogger(__file__)

RE_ML_LOG = re.compile(r'([\d\- 0-9:]+)(\w+)( \w+)(.*)')
RE_ERROR = re.compile(r'error', re.IGNORECASE)
RE_WARNING = re.compile(r'warning', re.IGNORECASE)


class ExecuteCommandError(Exception):

    def __init__(self, output, error):
        self.output = output
        self.error = error


def is_error(text):
    return RE_ERROR.search(text)


def is_warning(text):
    return RE_WARNING.search(text)


class Client(BungaClient):

    def __init__(self, uri, loop):
        super().__init__(uri)
        self._is_connected = False
        self._connected_event = asyncio.Event(loop=loop)
        self._complete_event = asyncio.Event(loop=loop)
        self._fout = None
        self._command_output = []
        self._awaiting_completion = False
        self._error = ''

    async def on_connected(self):
        self._is_connected = True
        self._connected_event.set()

    async def on_disconnected(self):
        self._connected_event.clear()
        self._is_connected = False

        if self._awaiting_completion:
            self._error = 'Connection lost'
            self._complete_event.set()

    def signal_completed(self, error):
        self._awaiting_completion = False
        self._error = error
        self._complete_event.set()

    async def on_execute_command_rsp(self, message):
        if message.output:
            self._command_output.append(message.output)
        else:
            self.signal_completed(message.error)

    async def on_log_entry_ind(self, message):
        pass

    async def on_get_file_rsp(self, message):
        if message.data:
            self._fout.write(message.data)
        else:
            self.signal_completed(message.error)

    async def on_put_file_rsp(self, message):
        self.signal_completed(message.error)

    async def wait_for_connection(self):
        if not self._is_connected:
            await self._connected_event.wait()

    async def execute_command(self, command):
        await self.wait_for_connection()
        self._awaiting_completion = True
        self._command_output = []
        message = self.init_execute_command_req()
        message.command = command
        self._complete_event.clear()
        self.send()
        await self._complete_event.wait()
        output = ''.join(self._command_output)

        if self._error:
            raise ExecuteCommandError(output, self._error)

        return output

    async def get_file(self, remote_path, local_path):
        await self.wait_for_connection()
        self._awaiting_completion = True

        with open(local_path, 'wb') as self._fout:
            message = self.init_get_file_req()
            message.path = remote_path
            self._complete_event.clear()
            self.send()
            await self._complete_event.wait()

        if self._error:
            raise Exception(self._error)

    async def put_file(self, local_path, remote_path):
        await self.wait_for_connection()
        self._awaiting_completion = True
        self._complete_event.clear()

        with open(local_path, 'rb') as fin:
            message = self.init_put_file_req()
            message.path = remote_path
            self.send()

            while True:
                message = self.init_put_file_req()
                message.data = fin.read(64)
                self.send()

                if not message.data:
                    break

        await self._complete_event.wait()

        if self._error:
            raise Exception(self._error)


def print_info(text):
    print(cyan(f'[bunga {time.strftime("%H:%M:%S")}] {text}', style='bold'))


def print_error(error):
    print(red(f'ERROR({error})', style='bold'))


def print_log_entry(entry, fout):
    header, text = entry.split(']', 1)
    header += ']'

    mo = RE_ML_LOG.match(text)

    if mo:
        date = mo.group(1)
        level = mo.group(2)
        text = level + mo.group(3) + mo.group(4)

        if level in ['EMERGENCY', 'ALERT', 'CRITICAL', 'ERROR']:
            text = red(text, style='bold')
        elif level == 'WARNING':
            text = yellow(text, style='bold')

        text = yellow(date) + text
    elif is_error(text):
        text = red(text, style='bold')
    elif is_warning(text):
        test = yellow(text, style='bold')

    print(green(header) + text, flush=True, file=fout)


class ClientThread(threading.Thread):

    def __init__(self, uri, client_class=Client):
        super().__init__()
        LOGGER.info('Server URI: %s', uri)
        self._loop = asyncio.new_event_loop()
        self._client = client_class(uri, self._loop)
        self.daemon = True

    def run(self):
        asyncio.set_event_loop(self._loop)
        self._loop.run_until_complete(self._start())
        self._loop.run_forever()

    def stop(self):
        pass

    def execute_command(self, command):
        return asyncio.run_coroutine_threadsafe(
            self._client.execute_command(command),
            self._loop).result()

    def get_file(self, remote_path, local_path):
        return asyncio.run_coroutine_threadsafe(
            self._client.get_file(remote_path, local_path),
            self._loop).result()

    def put_file(self, local_path, remote_path):
        return asyncio.run_coroutine_threadsafe(
            self._client.put_file(local_path, remote_path),
            self._loop).result()

    async def _start(self):
        self._client.start()


def create_to_path(from_path, to_path):
    if to_path:
        return to_path
    else:
        return os.path.basename(from_path)



import sys
import argparse
import logging

from .version import __version__
from .client import Client
from .client import ClientThread
from .client import ExecuteCommandError


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument('-l', '--log-level',
                        default='error',
                        choices=[
                            'debug', 'info', 'warning', 'error', 'critical'
                        ],
                        help='Set the logging level (default: %(default)s).')
    parser.add_argument('--version',
                        action='version',
                        version=__version__,
                        help='Print version information and exit.')

    # Workaround to make the subparser required in Python 3.
    subparsers = parser.add_subparsers(title='subcommands',
                                       dest='subcommand')
    subparsers.required = True

    # Import when used for less dependencies. For example, curses is
    # not part of all Python builds.
    from .subparsers import shell
    from .subparsers import get_file
    from .subparsers import put_file
    from .subparsers import log

    shell.add_subparser(subparsers)
    get_file.add_subparser(subparsers)
    put_file.add_subparser(subparsers)
    log.add_subparser(subparsers)

    args = parser.parse_args()

    level = logging.getLevelName(args.log_level.upper())
    logging.basicConfig(level=level, format='%(asctime)s %(message)s')

    if args.debug:
        args.func(args)
    else:
        try:
            args.func(args)
        except BaseException as e:
            sys.exit(f'error: {e}')



import sys
import logging
import asyncio
import unittest
from unittest.mock import patch
from io import StringIO

from colors import red
from colors import green
from colors import yellow

import bunga


def create_tcp_uri(listener):
    address, port = listener.sockets[0].getsockname()

    return f'tcp://{address}:{port}'


async def server_main(listener):
    async with listener:
        try:
            await listener.serve_forever()
        except asyncio.CancelledError:
            pass


class Client(bunga.Client):

    def __init__(self, listener):
        super().__init__(create_tcp_uri(listener), asyncio.get_event_loop())
        self.log_entries = asyncio.Queue()

    async def on_log_entry_ind(self, message):
        await super().on_log_entry_ind(message)
        await self.log_entries.put(''.join(message.text))


class ClientTest(unittest.TestCase):

    def test_execute_command(self):
        asyncio.run(self.execute_command())

    async def execute_command(self):
        async def on_client_connected(reader, writer):
            date_req = await reader.readexactly(12)
            self.assertEqual(date_req, b'\x01\x00\x00\x08\n\x06\n\x04date')
            writer.write(b'\x02\x00\x00\x08\x0a\x06\x0a\x04\x32\x30\x32\x30')
            writer.write(b'\x02\x00\x00\x02\x0a\x00')

            bad_req = await reader.readexactly(11)
            self.assertEqual(bad_req, b'\x01\x00\x00\x07\n\x05\n\x03bad')
            writer.write(b'\x02\x00\x00\x0d\x0a\x0b\x12\x09\x6e\x6f\x74\x20\x66'
                         b'\x6f\x75\x6e\x64')

            writer.close()

        listener = await asyncio.start_server(on_client_connected, 'localhost', 0)

        async def client_main():
            client = Client(listener)
            client.start()

            output = await client.execute_command('date')
            self.assertEqual(output, '2020')

            with self.assertRaises(bunga.ExecuteCommandError) as cm:
                await client.execute_command('bad')

            self.assertEqual(cm.exception.output, '')
            self.assertEqual(cm.exception.error, 'not found')

            client.stop()
            listener.close()

        await asyncio.wait_for(
            asyncio.gather(server_main(listener), client_main()), 2)

    def test_log_entry(self):
        asyncio.run(self.log_entry())

    async def log_entry(self):
        async def on_client_connected(reader, writer):
            writer.write(b'\x02\x00\x00\x0f\x12\r\n\x03123\n\x06Hello!')
            writer.write(b'\x02\x00\x00\x15\x12\x13\n\x03123\n\x06Hello!\n\x04 Hi!')
            writer.close()

        listener = await asyncio.start_server(on_client_connected, 'localhost', 0)

        async def client_main():
            client = Client(listener)
            client.start()
            log_entry = await client.log_entries.get()
            self.assertEqual(log_entry, '123Hello!')
            log_entry = await client.log_entries.get()
            self.assertEqual(log_entry, '123Hello! Hi!')
            client.stop()
            listener.close()

        await asyncio.wait_for(
            asyncio.gather(server_main(listener), client_main()), 2)

    def test_get_file(self):
        asyncio.run(self.get_file())

    async def get_file(self):
        async def on_client_connected(reader, writer):
            req = await reader.readexactly(13)
            self.assertEqual(req, b'\x01\x00\x00\x09\x12\x07\n\x05/init')
            writer.write(b'\x02\x00\x00\x0a\x1a\x08\x08\x08\x12\x041234')
            writer.write(b'\x02\x00\x00\x08\x1a\x06\x12\x045678')
            writer.write(b'\x02\x00\x00\x02\x1a\x00')

            writer.close()

        listener = await asyncio.start_server(on_client_connected, 'localhost', 0)

        async def client_main():
            client = Client(listener)
            client.start()

            await client.get_file('/init', 'get.txt')

            with open('get.txt', 'rb') as fin:
                self.assertEqual(fin.read(), b'12345678')

            client.stop()
            listener.close()

        await asyncio.wait_for(
            asyncio.gather(server_main(listener), client_main()), 2)

    def test_put_file(self):
        asyncio.run(self.put_file())

    async def put_file(self):
        async def on_client_connected(reader, writer):
            req = await reader.readexactly(22)
            self.assertEqual(
                req,
                b'\x01\x00\x00\x12\x1a\x10\n\x0eput_remote.txt')
            req = await reader.readexactly(17)
            self.assertEqual(req, b'\x01\x00\x00\r\x1a\x0b\x12\t12345678\n')
            req = await reader.readexactly(6)
            self.assertEqual(req, b'\x01\x00\x00\x02\x1a\x00')
            writer.write(b'\x02\x00\x00\x02"\x00')

            writer.close()

        listener = await asyncio.start_server(on_client_connected, 'localhost', 0)

        async def client_main():
            client = Client(listener)
            client.start()

            await client.put_file('tests/put.txt', 'put_remote.txt')

            client.stop()
            listener.close()

        await asyncio.wait_for(
            asyncio.gather(server_main(listener), client_main()), 2)

    def test_print_log_entry(self):
        stdout = StringIO()

        with patch('sys.stdout', stdout):
            bunga.client.print_log_entry(
                '[    0.141804] imx-sdma 20ec000.sdma: Direct firmware load for '
                'imx/sdma/sdma-imx6q.bin failed with error -2',
                sys.stdout)
            bunga.client.print_log_entry(
                '[    0.141826] imx-sdma 20ec000.sdma: external firmware not '
                'found, using ROM firmware',
                sys.stdout)
            bunga.client.print_log_entry(
                '[    2.497619] 1970-01-01 00:00:02 INFO dhcp_client State change '
                'from INIT to SELECTING.',
                sys.stdout)

        self.assertEqual(
            stdout.getvalue(),
            green('[    0.141804]')
            + red(' imx-sdma 20ec000.sdma: Direct firmware load for '
                  'imx/sdma/sdma-imx6q.bin failed with error -2', style='bold')
            + '\n'
            + green('[    0.141826]')
            + (' imx-sdma 20ec000.sdma: external firmware not found, using ROM '
               'firmware\n')
            + green('[    2.497619]')
            + yellow(' 1970-01-01 00:00:02 ')
            + ('INFO dhcp_client State change from INIT to SELECTING.\n'))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: bunga.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='bunga.proto',
  package='bunga',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x0b\x62unga.proto\x12\x05\x62unga\"\xab\x01\n\x0e\x43lientToServer\x12\x37\n\x13\x65xecute_command_req\x18\x01 \x01(\x0b\x32\x18.bunga.ExecuteCommandReqH\x00\x12)\n\x0cget_file_req\x18\x02 \x01(\x0b\x32\x11.bunga.GetFileReqH\x00\x12)\n\x0cput_file_req\x18\x03 \x01(\x0b\x32\x11.bunga.PutFileReqH\x00\x42\n\n\x08messages\"\xd8\x01\n\x0eServerToClient\x12\x37\n\x13\x65xecute_command_rsp\x18\x01 \x01(\x0b\x32\x18.bunga.ExecuteCommandRspH\x00\x12+\n\rlog_entry_ind\x18\x02 \x01(\x0b\x32\x12.bunga.LogEntryIndH\x00\x12)\n\x0cget_file_rsp\x18\x03 \x01(\x0b\x32\x11.bunga.GetFileRspH\x00\x12)\n\x0cput_file_rsp\x18\x04 \x01(\x0b\x32\x11.bunga.PutFileRspH\x00\x42\n\n\x08messages\"$\n\x11\x45xecuteCommandReq\x12\x0f\n\x07\x63ommand\x18\x01 \x01(\t\"2\n\x11\x45xecuteCommandRsp\x12\x0e\n\x06output\x18\x01 \x01(\t\x12\r\n\x05\x65rror\x18\x02 \x01(\t\"\x1b\n\x0bLogEntryInd\x12\x0c\n\x04text\x18\x01 \x03(\t\"\x1a\n\nGetFileReq\x12\x0c\n\x04path\x18\x01 \x01(\t\"7\n\nGetFileRsp\x12\x0c\n\x04size\x18\x01 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12\r\n\x05\x65rror\x18\x03 \x01(\t\"(\n\nPutFileReq\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\"\x1b\n\nPutFileRsp\x12\r\n\x05\x65rror\x18\x01 \x01(\tb\x06proto3')
)




_CLIENTTOSERVER = _descriptor.Descriptor(
  name='ClientToServer',
  full_name='bunga.ClientToServer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='execute_command_req', full_name='bunga.ClientToServer.execute_command_req', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='get_file_req', full_name='bunga.ClientToServer.get_file_req', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='put_file_req', full_name='bunga.ClientToServer.put_file_req', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='messages', full_name='bunga.ClientToServer.messages',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=23,
  serialized_end=194,
)


_SERVERTOCLIENT = _descriptor.Descriptor(
  name='ServerToClient',
  full_name='bunga.ServerToClient',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='execute_command_rsp', full_name='bunga.ServerToClient.execute_command_rsp', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='log_entry_ind', full_name='bunga.ServerToClient.log_entry_ind', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='get_file_rsp', full_name='bunga.ServerToClient.get_file_rsp', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='put_file_rsp', full_name='bunga.ServerToClient.put_file_rsp', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='messages', full_name='bunga.ServerToClient.messages',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=197,
  serialized_end=413,
)


_EXECUTECOMMANDREQ = _descriptor.Descriptor(
  name='ExecuteCommandReq',
  full_name='bunga.ExecuteCommandReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='command', full_name='bunga.ExecuteCommandReq.command', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=415,
  serialized_end=451,
)


_EXECUTECOMMANDRSP = _descriptor.Descriptor(
  name='ExecuteCommandRsp',
  full_name='bunga.ExecuteCommandRsp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='output', full_name='bunga.ExecuteCommandRsp.output', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='bunga.ExecuteCommandRsp.error', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=453,
  serialized_end=503,
)


_LOGENTRYIND = _descriptor.Descriptor(
  name='LogEntryInd',
  full_name='bunga.LogEntryInd',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='text', full_name='bunga.LogEntryInd.text', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=505,
  serialized_end=532,
)


_GETFILEREQ = _descriptor.Descriptor(
  name='GetFileReq',
  full_name='bunga.GetFileReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='path', full_name='bunga.GetFileReq.path', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=534,
  serialized_end=560,
)


_GETFILERSP = _descriptor.Descriptor(
  name='GetFileRsp',
  full_name='bunga.GetFileRsp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='size', full_name='bunga.GetFileRsp.size', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='bunga.GetFileRsp.data', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='bunga.GetFileRsp.error', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=562,
  serialized_end=617,
)


_PUTFILEREQ = _descriptor.Descriptor(
  name='PutFileReq',
  full_name='bunga.PutFileReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='path', full_name='bunga.PutFileReq.path', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='bunga.PutFileReq.data', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=619,
  serialized_end=659,
)


_PUTFILERSP = _descriptor.Descriptor(
  name='PutFileRsp',
  full_name='bunga.PutFileRsp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='error', full_name='bunga.PutFileRsp.error', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=661,
  serialized_end=688,
)

_CLIENTTOSERVER.fields_by_name['execute_command_req'].message_type = _EXECUTECOMMANDREQ
_CLIENTTOSERVER.fields_by_name['get_file_req'].message_type = _GETFILEREQ
_CLIENTTOSERVER.fields_by_name['put_file_req'].message_type = _PUTFILEREQ
_CLIENTTOSERVER.oneofs_by_name['messages'].fields.append(
  _CLIENTTOSERVER.fields_by_name['execute_command_req'])
_CLIENTTOSERVER.fields_by_name['execute_command_req'].containing_oneof = _CLIENTTOSERVER.oneofs_by_name['messages']
_CLIENTTOSERVER.oneofs_by_name['messages'].fields.append(
  _CLIENTTOSERVER.fields_by_name['get_file_req'])
_CLIENTTOSERVER.fields_by_name['get_file_req'].containing_oneof = _CLIENTTOSERVER.oneofs_by_name['messages']
_CLIENTTOSERVER.oneofs_by_name['messages'].fields.append(
  _CLIENTTOSERVER.fields_by_name['put_file_req'])
_CLIENTTOSERVER.fields_by_name['put_file_req'].containing_oneof = _CLIENTTOSERVER.oneofs_by_name['messages']
_SERVERTOCLIENT.fields_by_name['execute_command_rsp'].message_type = _EXECUTECOMMANDRSP
_SERVERTOCLIENT.fields_by_name['log_entry_ind'].message_type = _LOGENTRYIND
_SERVERTOCLIENT.fields_by_name['get_file_rsp'].message_type = _GETFILERSP
_SERVERTOCLIENT.fields_by_name['put_file_rsp'].message_type = _PUTFILERSP
_SERVERTOCLIENT.oneofs_by_name['messages'].fields.append(
  _SERVERTOCLIENT.fields_by_name['execute_command_rsp'])
_SERVERTOCLIENT.fields_by_name['execute_command_rsp'].containing_oneof = _SERVERTOCLIENT.oneofs_by_name['messages']
_SERVERTOCLIENT.oneofs_by_name['messages'].fields.append(
  _SERVERTOCLIENT.fields_by_name['log_entry_ind'])
_SERVERTOCLIENT.fields_by_name['log_entry_ind'].containing_oneof = _SERVERTOCLIENT.oneofs_by_name['messages']
_SERVERTOCLIENT.oneofs_by_name['messages'].fields.append(
  _SERVERTOCLIENT.fields_by_name['get_file_rsp'])
_SERVERTOCLIENT.fields_by_name['get_file_rsp'].containing_oneof = _SERVERTOCLIENT.oneofs_by_name['messages']
_SERVERTOCLIENT.oneofs_by_name['messages'].fields.append(
  _SERVERTOCLIENT.fields_by_name['put_file_rsp'])
_SERVERTOCLIENT.fields_by_name['put_file_rsp'].containing_oneof = _SERVERTOCLIENT.oneofs_by_name['messages']
DESCRIPTOR.message_types_by_name['ClientToServer'] = _CLIENTTOSERVER
DESCRIPTOR.message_types_by_name['ServerToClient'] = _SERVERTOCLIENT
DESCRIPTOR.message_types_by_name['ExecuteCommandReq'] = _EXECUTECOMMANDREQ
DESCRIPTOR.message_types_by_name['ExecuteCommandRsp'] = _EXECUTECOMMANDRSP
DESCRIPTOR.message_types_by_name['LogEntryInd'] = _LOGENTRYIND
DESCRIPTOR.message_types_by_name['GetFileReq'] = _GETFILEREQ
DESCRIPTOR.message_types_by_name['GetFileRsp'] = _GETFILERSP
DESCRIPTOR.message_types_by_name['PutFileReq'] = _PUTFILEREQ
DESCRIPTOR.message_types_by_name['PutFileRsp'] = _PUTFILERSP
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ClientToServer = _reflection.GeneratedProtocolMessageType('ClientToServer', (_message.Message,), dict(
  DESCRIPTOR = _CLIENTTOSERVER,
  __module__ = 'bunga_pb2'
  # @@protoc_insertion_point(class_scope:bunga.ClientToServer)
  ))
_sym_db.RegisterMessage(ClientToServer)

ServerToClient = _reflection.GeneratedProtocolMessageType('ServerToClient', (_message.Message,), dict(
  DESCRIPTOR = _SERVERTOCLIENT,
  __module__ = 'bunga_pb2'
  # @@protoc_insertion_point(class_scope:bunga.ServerToClient)
  ))
_sym_db.RegisterMessage(ServerToClient)

ExecuteCommandReq = _reflection.GeneratedProtocolMessageType('ExecuteCommandReq', (_message.Message,), dict(
  DESCRIPTOR = _EXECUTECOMMANDREQ,
  __module__ = 'bunga_pb2'
  # @@protoc_insertion_point(class_scope:bunga.ExecuteCommandReq)
  ))
_sym_db.RegisterMessage(ExecuteCommandReq)

ExecuteCommandRsp = _reflection.GeneratedProtocolMessageType('ExecuteCommandRsp', (_message.Message,), dict(
  DESCRIPTOR = _EXECUTECOMMANDRSP,
  __module__ = 'bunga_pb2'
  # @@protoc_insertion_point(class_scope:bunga.ExecuteCommandRsp)
  ))
_sym_db.RegisterMessage(ExecuteCommandRsp)

LogEntryInd = _reflection.GeneratedProtocolMessageType('LogEntryInd', (_message.Message,), dict(
  DESCRIPTOR = _LOGENTRYIND,
  __module__ = 'bunga_pb2'
  # @@protoc_insertion_point(class_scope:bunga.LogEntryInd)
  ))
_sym_db.RegisterMessage(LogEntryInd)

GetFileReq = _reflection.GeneratedProtocolMessageType('GetFileReq', (_message.Message,), dict(
  DESCRIPTOR = _GETFILEREQ,
  __module__ = 'bunga_pb2'
  # @@protoc_insertion_point(class_scope:bunga.GetFileReq)
  ))
_sym_db.RegisterMessage(GetFileReq)

GetFileRsp = _reflection.GeneratedProtocolMessageType('GetFileRsp', (_message.Message,), dict(
  DESCRIPTOR = _GETFILERSP,
  __module__ = 'bunga_pb2'
  # @@protoc_insertion_point(class_scope:bunga.GetFileRsp)
  ))
_sym_db.RegisterMessage(GetFileRsp)

PutFileReq = _reflection.GeneratedProtocolMessageType('PutFileReq', (_message.Message,), dict(
  DESCRIPTOR = _PUTFILEREQ,
  __module__ = 'bunga_pb2'
  # @@protoc_insertion_point(class_scope:bunga.PutFileReq)
  ))
_sym_db.RegisterMessage(PutFileReq)

PutFileRsp = _reflection.GeneratedProtocolMessageType('PutFileRsp', (_message.Message,), dict(
  DESCRIPTOR = _PUTFILERSP,
  __module__ = 'bunga_pb2'
  # @@protoc_insertion_point(class_scope:bunga.PutFileRsp)
  ))
_sym_db.RegisterMessage(PutFileRsp)


# @@protoc_insertion_point(module_scope)



import time

from ..client import Client
from ..client import ClientThread
from ..client import print_info
from ..client import print_log_entry


class LogClient(Client):

    async def on_connected(self):
        await super().on_connected()
        print_info('Connected')

    async def on_disconnected(self):
        await super().on_disconnected()
        print_info('Disconnected')

    async def on_log_entry_ind(self, message):
        print_log_entry(''.join(message.text))


def _do_log(args):
    client = ClientThread(args.uri, client_class=LogClient)
    client.start()

    while True:
        time.sleep(100)


def add_subparser(subparsers):
    subparser = subparsers.add_parser('log')
    subparser.add_argument('-u' ,'--uri',
                           default='tcp://127.0.0.1:28000',
                           help='URI of the server (default: %(default)s)')
    subparser.set_defaults(func=_do_log)



import threading
import sys
import socket
import bunga
import os
import unittest
from unittest.mock import patch
from io import StringIO


class ServerThread(threading.Thread):

    def __init__(self, listener, handler):
        super().__init__()
        self._listener = listener
        self._handler = handler
        self.daemon = True
        self.exception = None

    def run(self):
        try:
            self._handler(self._listener.accept()[0])
        except Exception as e:
            self.exception = e
            raise

        self._listener.close()


def start_server(handler):
    listener = socket.socket()
    listener.bind(('localhost', 0))
    listener.listen()
    server = ServerThread(listener, handler)
    server.start()

    return server, listener.getsockname()[1]


class CommandLineTest(unittest.TestCase):

    def shell_handler(self, client):
        req = client.recv(10)
        self.assertEqual(req, b'\x01\x00\x00\x06\n\x04\n\x02ko')
        client.sendall(b'\x02\x00\x00\x0d\n\x0b\x12\tNot found')
        req = client.recv(10)
        self.assertEqual(req, b'\x01\x00\x00\x06\n\x04\n\x02ls')
        client.sendall(b'\x02\x00\x00\x0f\n\r\n\x0bfoo bar fie')
        client.sendall(b'\x02\x00\x00\x02\n\x00')
        client.close()

    def test_shell(self):
        server, port = start_server(self.shell_handler)
        stdout = StringIO()

        with patch('sys.stdout', stdout):
            client = bunga.ClientThread(f'tcp://localhost:{port}',
                                        bunga.subparsers.shell.ShellClient)
            client.start()

            with self.assertRaises(bunga.ExecuteCommandError) as cm:
                client.execute_command('ko')

            self.assertEqual(str(cm.exception.error), 'Not found')

            output = client.execute_command('ls')
            self.assertIn('foo bar fie', output)
            client.stop()

        server.join()
        self.assertIsNone(server.exception)

    def get_file_handler(self, client):
        req = client.recv(21)
        self.assertEqual(req, b'\x01\x00\x00\x11\x12\x0f\n\rtests/put.txt')
        client.sendall(b'\x02\x00\x00\x0f\x1a\r\x08\t\x12\t12345678\n')
        client.sendall(b'\x02\x00\x00\x02\x1a\x00')
        client.close()

    def test_get_file(self):
        server, port = start_server(self.get_file_handler)
        argv = [
            'bunga', 'get_file',
            '--uri', f'tcp://localhost:{port}',
            'tests/put.txt'
        ]

        if os.path.exists('put.txt'):
            os.remove('put.txt')

        with patch('sys.argv', argv):
            bunga.main()

        with open('put.txt') as fin:
            self.assertEqual(fin.read(), '12345678\n')

        server.join()
        self.assertIsNone(server.exception)

    def get_file_error_handler(self, client):
        req = client.recv(21)
        self.assertEqual(req, b'\x01\x00\x00\x11\x12\x0f\n\rtests/put.txt')
        client.sendall(b'\x02\x00\x00\x0d\x1a\x0b\x1a\tNot found')
        client.close()

    def test_get_file_error(self):
        server, port = start_server(self.get_file_error_handler)
        argv = [
            'bunga', '-d', 'get_file',
            '--uri', f'tcp://localhost:{port}',
            'tests/put.txt'
        ]

        with patch('sys.argv', argv):
            with self.assertRaises(Exception) as cm:
                bunga.main()

            self.assertEqual(str(cm.exception), 'Not found')

        server.join()
        self.assertIsNone(server.exception)

    def put_file_handler(self, client):
        req = client.recv(15)
        self.assertEqual(req, b'\x01\x00\x00\x0b\x1a\t\n\x07put.txt')
        req = client.recv(17)
        self.assertEqual(req, b'\x01\x00\x00\r\x1a\x0b\x12\t12345678\n')
        req = client.recv(6)
        self.assertEqual(req, b'\x01\x00\x00\x02\x1a\x00')
        client.sendall(b'\x02\x00\x00\x02"\x00')
        client.close()

    def test_put_file(self):
        server, port = start_server(self.put_file_handler)
        argv = [
            'bunga', 'put_file',
            '--uri', f'tcp://localhost:{port}',
            'tests/put.txt'
        ]

        with patch('sys.argv', argv):
            bunga.main()

        server.join()
        self.assertIsNone(server.exception)

    def put_file_error_handler(self, client):
        req = client.recv(15)
        self.assertEqual(req, b'\x01\x00\x00\x0b\x1a\t\n\x07put.txt')
        req = client.recv(17)
        self.assertEqual(req, b'\x01\x00\x00\r\x1a\x0b\x12\t12345678\n')
        req = client.recv(6)
        self.assertEqual(req, b'\x01\x00\x00\x02\x1a\x00')
        client.sendall(b'\x02\x00\x00\x0d"\x0b\n\tNot found')
        client.close()

    def test_put_file_error(self):
        server, port = start_server(self.put_file_error_handler)
        argv = [
            'bunga', '-d', 'put_file',
            '--uri', f'tcp://localhost:{port}',
            'tests/put.txt'
        ]

        with patch('sys.argv', argv):
            with self.assertRaises(Exception) as cm:
                bunga.main()

            self.assertEqual(str(cm.exception), 'Not found')

        server.join()
        self.assertIsNone(server.exception)



from ..client import ClientThread
from ..client import create_to_path


def _do_get_file(args):
    client = ClientThread(args.uri)
    client.start()
    localfile = create_to_path(args.remotefile, args.localfile)
    client.get_file(args.remotefile, localfile)


def add_subparser(subparsers):
    subparser = subparsers.add_parser('get_file')
    subparser.add_argument('-u' ,'--uri',
                           default='tcp://127.0.0.1:28000',
                           help='URI of the server (default: %(default)s)')
    subparser.add_argument('remotefile', help='The remote file path.')
    subparser.add_argument('localfile',
                           nargs='?',
                           help='The local file path.')
    subparser.set_defaults(func=_do_get_file)



import sys
import os
import subprocess

from prompt_toolkit import prompt
from prompt_toolkit.history import FileHistory
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory

from ..client import Client
from ..client import ClientThread
from ..client import print_info
from ..client import print_error
from ..client import print_log_entry
from ..client import ExecuteCommandError


class ShellClient(Client):

    async def on_connected(self):
        await super().on_connected()
        print_info('Connected')

    async def on_disconnected(self):
        await super().on_disconnected()
        print_info('Disconnected')


def is_comment(line):
    """Lines starting with "#" are comments.

    """

    return line.strip().startswith('#')


def fprint_output(command, output, fout):
    if command.startswith('dmesg'):
        for line in output.splitlines():
            print_log_entry(line, fout)
    else:
        print(output, end='', flush=True, file=fout)


def print_output(command, pipe_commands, output):
    if pipe_commands:
        with subprocess.Popen(pipe_commands,
                              shell=True,
                              stdin=subprocess.PIPE,
                              encoding='utf-8') as proc:
            fprint_output(command, output, proc.stdin)
    else:
        fprint_output(command, output, sys.stdout)


def shell_main(client):
    commands = [command[1:] for command in []]
    commands.append('exit')
    user_home = os.path.expanduser('~')
    history = FileHistory(os.path.join(user_home, '.bunga-history.txt'))

    while True:
        try:
            line = prompt('$ ',
                          complete_while_typing=True,
                          auto_suggest=AutoSuggestFromHistory(),
                          enable_history_search=True,
                          history=history)
        except EOFError:
            break

        if line:
            if is_comment(line):
                continue

            if line == 'exit':
                break

            pipe_command, _, pipe_commands = line.partition('|')
            redirect_command, _, redirect_commands = line.partition('>')

            if len(pipe_command) <= len(redirect_command):
                command = pipe_command
            else:
                command = redirect_command
                pipe_commands = f'cat > {redirect_commands}'

            try:
                output = client.execute_command(command)
                print_output(command, pipe_commands, output)
            except ExecuteCommandError as e:
                print(e.output, end='')
                print_error(e.error)


def _do_shell(args):
    client = ClientThread(args.uri, ShellClient)
    client.start()

    shell_main(client)


def add_subparser(subparsers):
    subparser = subparsers.add_parser('shell')
    subparser.add_argument('-u' ,'--uri',
                           default='tcp://127.0.0.1:28000',
                           help='URI of the server (default: %(default)s)')
    subparser.set_defaults(func=_do_shell)
